unit IntUnit;

interface
uses classes,sysutils,intlists,aconsts,forms,windows,dialogs;
{----------------------------------------------}
type  tbuf=array[1..12000] of char;
      pbuf=^tbuf;

type THalCompiler=class
     public
       AllowPrint:Boolean;
       DolEnabled:Boolean;
       HalOwner:TObject;
       IdentPrefix:String;
       Token:TToken;
       spoint:integer;
       inbuf:^tbuf;
       fout:tstream;
       curread,numread:integer;
       Labels,Variables:TIdentList;
       Consts:TConstList;
       IDLabels:TIDLabelList;
       Prog:TProgCollect;
       lastvar:TIdentListItem;
       lastobject:TObject;
       mobjects:tobjcollect;
       notfromobj:boolean;
       FLastClassType:String;
       nextstate:boolean;
       function  NextByte:byte;
       function  getinternalfun(const sname:string;ident:tidentlistitem):integer;
       function  VarIDByName(Const AName:String;var Id:Integer):boolean;
       procedure getprocbody(i:integer);
       procedure getvartype;
       procedure getafterproc;
       function  getprocdef(askip:boolean):Integer;
       procedure getclassdefblock(b:boolean);
       procedure getmyvariables(b:boolean);
       function  ReadByte:byte;
       procedure findobject(var s:string);
       procedure gettypeidentifier;
       procedure BackBytes(a:integer);
       function  ReadToken:TToken;
       function  getvar:integer;
       Function  NextToken:TToken;
       procedure putcode(aid,a,b:integer);
       procedure getdelimeter;
       function  iexpression:integer;
       function  isimpleexpression:integer;
       Function  GetOpenArray(NeedSq:boolean):Integer;
       procedure getuserfunction(const name:string;scope,fik:boolean);
       procedure getcoma;
       procedure getidentifier;
       function  iterm:integer;
       function  ifactor:integer;
       procedure getclosebracket;
       procedure getopenbracket;
       function  VarByName(Const AName:String;var Ident:TIdentListItem):boolean;
       function  AddVars(const Aname:string;ID:Integer):Integer;
       function  myexpress:integer;
       procedure equaldispath(const VarName:String);
       procedure labeldispatch(const labelname:string);
       procedure getprogramname;
       procedure getclassdef(const cname:string);
       procedure getdeclarations;
       procedure myinit;
       procedure mydone;
       procedure getoperatorcoma;
       procedure getwhileoperator;
       procedure getrepeatoperator;
       procedure getforoperator;
       procedure getifoperator;
       procedure getoperator;
       procedure getoperatorblock;
       procedure compileprogram;
       procedure writeprogram;
       procedure error(const s:string);
       constructor Create(afin:TMemoryStream;afout:TStream);
       procedure compile;virtual;
       destructor Destroy;override;
     end;
{------------------------------}
type EBinError=class(exception);

     TBinProg=Class
     public
      curpos:integer;
      CallStack:Array[0..1000] of integer;
      CallMax:Integer;
      stacksize:integer;
      ConstSize:integer;
      ProgSize :integer;
      LabelSize:integer;
      Loaded   :boolean;
      c:pvariantarray;
      s:pvariantarray;
      prog:pprogarray;
      l:pintarray;
      lastobject:tobject;
      function  dispatch(opcode,a,b:integer):boolean;
      procedure RunFrom(Acurpos:Integer);
      constructor Create(Fin:TStream);
      procedure   Run;
      destructor  Destroy;override;
      procedure   Error(const s:string);
     end;

     THalRuner=class
     private
     instream:tmemorystream;
     Compiled:boolean;
     mymem:TMemoryStream;
     public
     HalOwner:TObject;
     myobjects:tobjcollect;
     FBinProg:TBinProg;
     loadtemp:boolean;  { set true to copy Variables to TempVarList}
     dontrun:boolean;  { set true if you need to compile only}
     TempVarList:TIdentList;
     FLastClassType:String;
     procedure AddObject(const objname,objtype:string;tobj:tobject);
     procedure AddObjectbyRef(const objname:string;tobj:tobject);
     constructor Create(T:TMemoryStream);
     procedure Run;
     destructor  Destroy;override;
     end;
{----------------------------------------------}
implementation
{----------------------------------------------}
procedure THalCompiler.myinit;
begin
  Labels:=TIdentList.Create;
  variables:=TIdentList.Create;
  consts:=TConstList.Create(tconstitem);
  IDLabels:=TIDLabelList.Create(tidlabelitem);
  Prog:=TProgCollect.Create(tprogitem);
end;
{----------------------------------------------}
procedure THalCompiler.mydone;
begin
  Variables.Free;
  Consts.Free;
  Labels.Free;
  IDLabels.Free;
  Prog.Free;
end;
{----------------------------------------}
procedure THalCompiler.Error(const s:string);
var
  i:integer;
  linenumber:integer;
  sa:string;
begin
  linenumber:=1;
  for i:=1 to curread do if inbuf[i]=char(idnewline) then inc(linenumber);
  sa:=FmtLoadStr(synt_err,[InttoStr(linenumber),#13#10+s]);
  If AllowPrint then myerrorout(sa);
  raise ECompilerError.Create(sa);
end;
{------------------------------}
function THalCompiler.ReadByte:byte;
begin
  inc(curread);
  if curread>numread then begin Result:=idEndOfFile;exit;end;
  Result:=byte(Inbuf[curread]);
end;
{------------------------------}
function THalCompiler.NextByte:byte;
var i:integer;
begin
 i:=curread;
 Result:=ReadByte;
 curread:=i;
end;
{------------------------------}
procedure THalCompiler.BackBytes(a:integer);
begin
 if curread-a>=0 then dec(curread,a);
end;
{------------------------------}
function THalCompiler.ReadToken:TToken;
var b,begst,a:byte;
    s:string;
    i:integer;
label label2,label1,l9,l8,l10,l11;

begin
 Application.ProcessMessages;
label1:
a:=ReadByte;
if char(a) in BlackSpaces then goto label1;
{--}
if a=idslash then begin
if ReadByte<>idslash then begin backbytes(1);a:=idslash;goto l9;end;
l8:a:=readbyte;
   if a=idEndOFFile then goto l9;
   if not(a<32) then goto l8;
   goto label1;
end;
l9:
{-----}
if a=idminus then begin
if ReadByte<>idminus then begin backbytes(1);a:=idminus;goto l11;end;
s:='';
l10:a:=readbyte;
   if a=idEndOFFile then goto l11;
   if not(a<32) then begin s:=s+Char(a);goto l10;end else
   If (length(S)>0) and (nextstate=false) and (AllowPrint) then MyStdOut(s);
   goto label1;
end;
l11:
{-----}
if ((a=idopenbracket) or (a=idslash)) and (nextbyte=idstar) then begin
 b:=a;
 repeat
  a:=ReadByte;
  case a of
   idstar:if nextbyte=b then begin
           a:=readbyte;
           goto label1;
          end;
   idEndOfFile:Error(LoadStr(eofincom_ERR));
  end;
 until false;
end;
{----}
if char(a) in WhiteSpaces then
 begin
  DolEnabled:=(a=iddollar);
  Result:=SetToken(a,0);
  exit;
 end;
case a of
idOpenComment:begin
               repeat
                a:=ReadByte;
                case a of
                 idCloseComment:goto label1;
                 idEndOfFile:Error(LoadStr(eofincom_ERR));
                end;
               until false;
              end;
IdStringChar,Id2StringChar: begin
              s:='';begst:=a;
              repeat
              a:=ReadByte;
              if a=begst then begin Result:=SetToken(idStringConst,s);exit;end
              else
              if a=idEndOfFile then
              Error(LoadStr(eofinstring_ERR))
              else s:=s+char(a);
              until false;
              end;
idEndOFFile:begin Result:=SetToken(idEndOFFile,0);exit;end;
end;
S:='';
label2:
S:=S+char(a);
a:=ReadByte;
if not(char(a) in StopChars) then goto label2;
BackBytes(1);
s:=AnsiUpperCase(s);
if s[1] in FirstIdentChar then begin
 if length(s)>1 then
 for i:=2 to length(s) do
   if not(s[i] in identbackchars) then Error(FmtLoadStr(bad_idName,[s]));
 if ResWords.IDByName(s,i) then  Result:=SetToken(i,s)
 else if ResConsts.Find(s,i) then Result:=SetToken(idresconst,i) else
 Result:=SetToken(idIdentifier,s);
end else
 begin
  If s[2]='X' then begin  {need checking}
   Result:=SetToken(idHexConst,copy(s,3,length(s)));
  end else begin
  for i:=1 to length(s) do
  if not((s[i] in Digit) or ((DolEnabled) and (S[i] in HExDIgit)))
   then Error(FmtLoadStr(bad_id,[s]));
  Result:=SetToken(idNumberConst,s);
  end;
 end;
end;
{------------------------------}
Function THalCompiler.NextToken:TToken;
var i:integer;
begin
 nextstate:=true;
 i:=curread;
 Result:=ReadToken;
 curread:=i;
 nextstate:=false;
end;
{----------------------------------------}
procedure THalCompiler.putcode(aid,a,b:integer);
begin prog.putop(aid,a,b);end;
{------------------------------}
procedure THalCompiler.getdelimeter;
var Token:TToken;
begin Token:=ReadToken;
 if (Token.id<>iddelimeter) then Error(LoadStr(delimeter_expected));
end;
{------------------------------}
function THalCompiler.iexpression:integer;
var i:integer;
begin
Result:=isimpleexpression;
case nexttoken.id of
{ = }
idequal  : begin
           token:=readtoken;
           i:=isimpleexpression;
           putcode(ocequal,Result,i);
           end;
{ > >= }
idgreater: begin
           token:=readtoken;
           if nexttoken.id=idequal then begin
           token:=readtoken;
           i:=isimpleexpression;
           putcode(ocgreaterequal,Result,i);
           end else
           begin
           i:=isimpleexpression;
           putcode(ocgreater,Result,i);
           end;
           end;
{ < <= <> }
idless   : begin
           token:=readtoken;
           case nexttoken.id of
           idequal: begin
                    token:=readtoken;
                    i:=isimpleexpression;
                    putcode(oclessequal,Result,i);
                    end;
           idgreater:begin
                     token:=readtoken;
                     i:=isimpleexpression;
                     putcode(ocnotequal,Result,i);
                     end;
           else begin
                i:=isimpleexpression;
                putcode(ocless,Result,i);
                end;
           end;
           end;
end;
end;
{------------------------------}
function THalCompiler.isimpleexpression:integer;
label l1;
var i:integer;
begin
Result:=iterm;
l1:
case nexttoken.id of
idPlus: begin Token:=ReadToken;
        i:=iterm;
        putcode(ocadd,Result,i);
        goto l1;
        end;
idMinus:begin
        Token:=ReadToken;
        i:=iterm;
        putcode(ocsub,Result,i);
        goto l1;
        end;
id_Or: begin Token:=ReadToken;
        i:=iterm;
        putcode(ocor,Result,i);
        goto l1;
        end;
id_Xor: begin Token:=ReadToken;
        i:=iterm;
        putcode(ocxor,Result,i);
        goto l1;
        end;
end;
end;
{------------------------------}
function THalCompiler.iterm:integer;
var i:integer;
label l1;
begin
Result:=ifactor;
l1:
case nexttoken.id of
id_shl: begin Token:=ReadToken;
        i:=ifactor;
        putcode(ocshl,Result,i);
        goto l1;
        end;
id_shr:  begin Token:=ReadToken;
        i:=ifactor;
        putcode(ocshr,Result,i);
        goto l1;
        end;
id_And: begin Token:=ReadToken;
        i:=ifactor;
        putcode(ocand,Result,i);
        goto l1;
        end;
idStar: begin Token:=ReadToken;
        i:=ifactor;
        putcode(ocmul,Result,i);
        goto l1;
        end;
idSlash:begin
        Token:=ReadToken;
        i:=ifactor;
        putcode(ocslash,Result,i);
        goto l1;
        end;
id_div:begin
        Token:=ReadToken;
        i:=ifactor;
        putcode(ocdiv,Result,i);
        goto l1;
        end;
id_mod:begin
        Token:=ReadToken;
        i:=ifactor;
        putcode(ocmod,Result,i);
        goto l1;
        end;
end;
end;
{------------------------------}
procedure THalCompiler.getopenbracket;
begin
token:=ReadToken;
if (token.id<>idOpenBracket) and (token.id<>idsqOpenBracket)
 then Error(loadstr(need_opbr));
end;
{------------------------------}
procedure THalCompiler.getclosebracket;
begin
token:=ReadToken;
if (token.id<>idcloseBracket) and (token.id<>idsqcloseBracket)
 then Error(LoadStr(need_clbr));
end;
{------------------------------}
procedure THalCompiler.getcoma;
begin
 token:=readtoken;
 if token.id<>idcomma then Error(LoadStr(comma_expected));
end;
{------------------------------}
{ If only 1 parameter of open array type you don't need [] }
Function THalCompiler.GetOpenArray(NeedSq:boolean):Integer;
var ints:Array[0..200] of integer;
    i,maxints:integer;
    ospoint:integer;
begin
 If (NeedSq) or (nexttoken.id=idsqopenbracket) then begin
  token:=readtoken;
  If token.id<>idsqopenbracket then Error(FmtLoadStr(opsq_exp,[token.data]));
 end;
{----}
 maxints:=0;
 ospoint:=spoint;
 inc(spoint);
 while (nexttoken.id<>idsqclosebracket) and
 (nexttoken.id<>idclosebracket) do begin
  IF maxints>0 then getcoma;
  ints[maxints]:=iexpression;
  inc(maxints);
 end;
 putcode(ocvarraycreate,ospoint,maxints);

 putcode(ocloadconst,spoint,Consts.newitem(maxints));

 putcode(ocsetvarray,0,spoint);
 inc(spoint);

 putcode(0,ospoint,0);
 For i:=0 to maxints-1 do begin
  putcode(ocsetvarray,i+1,ints[i]);
  putcode(0,ospoint,0);
 end;
 Result:=ospoint;
{----}
 If (NeedSq) or (nexttoken.id=idsqclosebracket) then begin
  token:=readtoken;
  If token.id<>idsqclosebracket then
   Error(FmtLoadStr(clsq_exp,[token.data]));
 end;
end;
{------------------------------}
procedure THalCompiler.getuserfunction(const name:string;scope,fik:boolean);
var oldstack,i,index:integer;
    f:TFunListItem;
    ia:array[0..maxparams] of integer;
    sname,sname1:string;
    ident:tidentlistitem;
    olastobj:TObject;
    onotfromobj:boolean;
label lt,lr,lr1;
begin

 olastobj:=lastobject;onotfromobj:=notfromobj;
if varbyname(name,ident) then begin
 i:=getinternalfun(name,ident);
 putcode(ocmov,spoint,i);
 exit;
end;

if (not assigned(Funs)) then Error(FmtLoadStr(unkn_id,[name]));


If Fik then begin
 sname:=name+'_VET';
 while (Funs.Find(sname,index)=false)
 do begin
 i:=pos('.',sname);
 if i=0 then goto lr1;
 sname1:=GetPearent(copy(sname,1,i-1));
 if sname1='' then goto lr1;
 delete(sname,1,i-1);
 sname:=sname1+sname;
 end;
goto lr;
end;
lr1:sname:=name;
while (Funs.Find(sname,index)=false)
do begin
i:=pos('.',sname);
if i=0 then Error(FmtLoadStr(unkn_id,[name]));
sname1:=GetPearent(copy(sname,1,i-1));
if sname1='' then Error(FmtLoadStr(unkn_id,[name]));
delete(sname,1,i-1);
sname:=sname1+sname;
end;
lr:
f:=TFunListItem(Funs.Objects[index]);
if f.parcount>0 then getopenbracket;
if (f.fun<>scope) and (f.fun) then
begin
if scope then Error(LoadStr(in_funuse)) else Error(LoadStr(in_procuse));
end;
if f.parcount>0 then begin
 oldstack:=spoint;
 for i:=0 to F.Parcount-1 do begin
  case f.Params[i] of
   0:ia[i]:=iexpression;
   1:ia[i]:=getvar;
   3:ia[i]:=getopenarray(F.ParCount<>1);
   4:begin
      getclosebracket;
      token:=readtoken;
      if token.id<>id2points then Error(LoadStr(p2_exp));
      token:=readtoken;
      if token.id<>idequal then Error(LoadStr(eq_exp));
      ia[i]:=iexpression;
      goto lt;
     end;
   else Error(LoadStr(unsup_partype));
  end;
 if (i<f.parcount-1) and (f.params[i+1]<>4) then getcoma;
 end;
 lt:
 spoint:=oldstack;
end;
if (olastobj<>nil) or (onotfromobj) then begin
 If oNotFromObj then putcode(ocselffromvar,Integer(olastobj),0) else
 putcode(ocsetself,integer(olastobj),0);
end;
putcode(ocextfun,spoint,index);
if f.parcount>0 then begin
i:=0;
while i<=f.parcount-1 do begin
putcode(ia[i],ia[i+1],ia[i+2]);
inc(i,3);
end;
end;
if (f.parcount>0) and (F.Params[f.parcount-1]<>4) then getclosebracket;
end;
{------------------------------}
function THalCompiler.AddVars(const Aname:string;ID:Integer):Integer;
begin
 Result:=Variables.additem(IdentPrefix+Aname,ID);
end;
{------------------------------}
function THalCompiler.VarIDByName(Const AName:String;var Id:Integer):boolean;
var S:String;
    i:integer;
label l1;
begin
 S:=IdentPrefix;
l1:
 Result:=Variables.Find(s+Aname,id);
 if (result) or (s='') then exit;
 i:=length(s);
 If S[i]='.' then dec(i);
 while (S[i]<>'.') and (i>0) do dec(i);
 setlength(s,i);
 goto l1;
end;
{------------------------------}
function THalCompiler.VarByName(Const AName:String;var Ident:TIdentListItem):boolean;
var i:integer;
begin
 Result:=VarIDByName(aname,i);
 if result then ident:=(Variables.Objects[i] as tidentlistitem)
  else ident:=nil;
end;
{------------------------------}
procedure THalCompiler.findobject(var s:string);
var tob:tobjitem;
    ident:tIdentListItem;
begin
 notfromobj:=false;
if mobjects.objbyname(s,tob) then
  begin lastobject:=tob.tobj;s:=tob.objtype;
        UseUnit(s);
  end
else
if (varbyname(s,ident)) and (RegisteredClassName(ident.vtype))
then
  begin
  notfromobj:=true;lastobject:=TObject(ident.id);
  s:=ident.vtype;
  end
else lastobject:=nil;
end;
{------------------------------}
function THalCompiler.getinternalfun(
         const sname:string;ident:tidentlistitem):integer;
var i,j:integer;
    pind:array[0..100] of integer;
    im:tidentlistitem;
    temp:string;
begin
 with ident do begin
  if parcount>0 then begin
   getopenbracket;
   for i:=0 to parcount-1 do begin
    if params[i]=0 then pind[i]:=iexpression
    else pind[i]:=getvar;

   temp:=sname+'.'+paramnames.strings[i];
   if varbyname(temp,im)=false then Error(FmtLoadStr(proc_notfound,[temp]));

    putcode(ocmov,im.id,pind[i]);
    if i<parcount-1 then getcoma;
   end;
   getclosebracket;
  end;

  if varbyname(sname+coprocresult,im) then Result:=im.id
   else Error(FmtLoadStr(no_resvar,[sname]));
  if varbyname(sname,im) then putcode(occall,im.id,0)
   else error(FmtLoadStr(proc_notfound,[sname]));

  for i:=0 to parcount-1 do if params[i]<>0 then begin
   varbyname(sname+'.'+paramnames.strings[i],im);
   putcode(ocmov,pind[i],im.id);
  end;
 end;
 end;
{------------------------------}
function THalCompiler.ifactor:integer;
var i,v,a2,a1,a,code,l:integer;
    mdouble:double;
    s:string;
    funid,varid:integer;
    ident:TIdentListItem;
    mextended:extended;
    tob:tobjitem;
label label1,lab1,lop;
begin
Token:=ReadToken;l:=token.id;
case l of
idsqopenbracket:begin          {!!!! set but not tested}
                 Result:=GetOpenArray(false);
                 token:=readtoken;
                 If token.id<>idsqclosebracket then
                 Error(FmtLoadStr(clsq_exp,[token.data]));
                end;
ididentifier :begin
              s:=token.data;
              if nexttoken.id=idpoint then begin
               findobject(s);
              end;
              while nexttoken.id=idpoint do begin
              token:=readtoken;
              token:=readtoken;
              if token.id<>ididentifier then Error(LoadStr(id_expected));
              s:=s+'.'+token.data;
              end;
              if (nexttoken.id=idopenbracket) or (nexttoken.id=idsqopenbracket)
              then begin
              lab1:
              getuserfunction(s,false,false);
              Result:=spoint;
              inc(spoint);
              end else
              begin

               {--}
              If Mobjects.ObjbyName(s,tob) then begin
               putcode(ocloadconst,spoint,Consts.newitem(Integer(tob.tobj)));
               Result:=spoint;
               inc(spoint);
               end else
               if DynaVars.GetDynaObject(s,HalOwner,varid,funid) then
               begin
                putcode(ocloadextvar,varid,funid);
                putcode(ocbackcode,spoint,0);
                Result:=spoint;
                inc(spoint);
               end else
               begin
               if varbyname(s,ident) then begin
                case ident.identtype of
                 tivariable: begin
                              putcode(ocmov,spoint,ident.id);
                              Result:=spoint;
                              inc(Spoint);
                             end;
                else Error(LoadStr(in_procuse));
                end;
               end else goto lab1;
               end;
               {--}
              end;

              end;
idresconst:begin
            putcode(ocloadconst,spoint,Consts.newitem(
            TResConstListItem(ResConsts.Objects[token.data]).Value
            ));
            Result:=spoint;
            inc(spoint);
           end;
id_false: begin
          putcode(ocloadconst,spoint,Consts.newitem(FALSE));
          Result:=spoint;
          inc(spoint);
          end;
{id_format:begin getformatfunction;Result:=spoint;inc(spoint);end;}
id_true: begin
          putcode(ocloadconst,spoint,Consts.newitem(TRUE));
          Result:=spoint;
          inc(spoint);
         end;
id_nil: begin
          putcode(ocloadconst,spoint,Consts.newitem(0));
          Result:=spoint;
          inc(spoint);
         end;
iddollar:begin
         token:=readtoken;
          {IFDEF DEBUG}
          {Showmessage(token.data+':'+IntToSTr(token.id));}
          {ENDIF}
         if token.id<>idnumberconst then
          Error(LoadStr(bad_hex));
         lop:
         v:=Hex2Dec(token.data);
         putcode(ocloadconst,spoint,Consts.newitem(v));
         Result:=spoint;
         inc(spoint);
         end;
idhexconst:begin
           goto lop;
           end;
idnumberconst:begin
               if nexttoken.id=idpoint then begin
               s:=token.data;
               token:=readtoken;
               token:=readtoken;
               if token.id<>idnumberconst then
               Error(LoadStr(bad_realconst));
               s:=s+'.'+token.data;
               val(s,mextended,code);
               putcode(ocloadconst,spoint,Consts.newitem(mextended));
               end else begin
               val(token.data,v,code);
               putcode(ocloadconst,spoint,Consts.newitem(v));
               end;
               Result:=spoint;
               inc(spoint);
              end;
idnotequal:  begin
               token:=readtoken;
               if token.id<>idnumberconst then
               Error(loadStr(bad_charconst));
               val(token.data,v,code);
               putcode(ocloadconst,spoint,Consts.newitem(Char(v)));
               Result:=spoint;
               inc(spoint);
               end;
idstringconst:
               begin
               S:=token.data;
               label1:
               case nexttoken.id of
               idnotequal:begin token:=readtoken;
                         token:=readtoken;
                         if token.id<>idnumberconst then
                         Error(loadStr(bad_charconst));
                         val(token.data,v,code);
                         s:=s+char(v);
                         goto label1;
                         end;
               idstringconst:begin token:=readtoken;S:=S+token.data;
                                   goto label1;
                                   end;
               end;
               putcode(ocloadconst,spoint,Consts.newitem(s));
               Result:=spoint;
               inc(spoint);
               end;
idopenbracket: begin
               Result:=iexpression;
               getclosebracket;
               end;
idplus       : begin
               Result:=ifactor;
               end;
idminus      : begin
               Result:=ifactor;
               putcode(ocneg,Result,0);
               end;
id_not       : begin
               Result:=ifactor;
               putcode(ocnot,Result,0);
               end;
end;
end;
{------------------------------------}
{$ifdef Demo}
function DelphiRunning : boolean;
var
  H1, H2, H3, H4 : Hwnd;
const
  A1 : array[0..12] of char = 'TApplication'#0;
  A2 : array[0..15] of char = 'TAlignPalette'#0;
  A3 : array[0..18] of char = 'TPropertyInspector'#0;
  A4 : array[0..11] of char = 'TAppBuilder'#0;
  T1 : array[0..6] of char = 'Delphi'#0;
begin
  H2 := FindWindow(A2, nil);
  H3 := FindWindow(A3, nil);
  H4 := FindWindow(A4, nil);
  Result := (H2 <> 0) and
            (H3 <> 0) and (H4 <> 0);
end;
{$endif}
{------------------------------}
function THalCompiler.myexpress:integer;
var i:integer;
begin
i:=spoint;
Result:=iexpression;
spoint:=i;
end;
{------------------------------}
procedure THalCompiler.equaldispath(const VarName:String);
var i,j:integer;
    index:integer;
    f:tfunlistitem;
    vname:string;
{-----------------------------}
function getobjfuns(var idx:integer):boolean;
var i:integer;
    sname1:string;
begin
Result:=false;
while Funs.Find(vname+'_SET',idx)=false do begin
i:=pos('.',vname);if i=0 then exit;
sname1:=GetPearent(copy(vname,1,i-1));if sname1='' then exit;
delete(vname,1,i-1);
vname:=sname1+vname;
end;
Result:=True;
end;
{-----------------------------}
var varid,funid:integer;
    Ident:TIdentListItem;
    olastobj:tobject;
    onotfromobj:boolean;
begin

vname:=varname;
if (varbyname(vname,ident)) and (ident.identtype=0) then
   begin
   i:=myexpress;
   j:=ident.id;
   putcode(ocmov,j,i);
   end
else
if DynaVars.GetDynaObject(vname,HalOwner,varid,funid) then
   begin
   i:=myexpress;
   putcode(ocstoreextvar,varid,funid);
   putcode(ocbackcode,i,0);
   end
else
if (assigned(Funs)) and (pos('.',vname)<>0) and (getobjfuns(index)) then
   begin
    olastobj:=lastobject;onotfromobj:=notfromobj;
    i:=myexpress;
    f:=TFunListItem(Funs.Objects[index]);
    if (olastobj<>nil) or (onotfromobj) then begin
      If oNotFromObj then putcode(ocselffromvar,Integer(olastobj),0) else
      putcode(ocsetself,integer(olastobj),0);
    end;

    putcode(ocextfun,spoint,index);
    putcode(i,0,0);
   end
else Error(FmtLoadStr(var_NotDef,[varname]));
end;
{------------------------------}
procedure THalCompiler.labeldispatch(const labelname:string);
var pl,i:integer;
begin
if Labels.IDByName(labelname,i)=false then
Error(FmtLoadStr(lab_notdef,[labelname]));

pl:=prog.count;
IDLabels.setplace(i,pl);
end;
{------------------------------}
procedure THalCompiler.getprogramname;
var i:integer;
begin
i:=nexttoken.id;
If (i=id_Program) or (i=id_unit) then
begin
Token:=ReadToken;
Token:=ReadToken;
if Token.Id<>idIdentifier then
Error(LoadStr(progname_exp));
getdelimeter;
end;end;
{------------------------------}
procedure THalCompiler.getidentifier;
begin
 token:=readtoken;
 if token.id<>ididentifier then Error(LoadStr(id_expected));
end;
{------------------------------}
{ askip not supported
  without delimeter}
function THalCompiler.getprocdef(askip:boolean):Integer;
var ProcName:String;
    k,AParamCount:Integer;
    AParams:tbytearray;
    MAdd:Boolean;
    ident:integer;
    EResult:TIdentListItem;
{--}
procedure getprocparams;
label l1,l2;
var k,curtp,st:integer;
    begv:integer;
    il:array[1..100] of integer;
    sname:string;
begin
l1:curtp:=0;st:=0;
 case nexttoken.id of
  ididentifier  :begin
                 l2:
                  getidentifier;
                  if madd=false then begin
                   inc(st);
                   sname:=ProcName+'.'+Token.Data;
                   il[st]:=AddVars(sname,spoint);
                   EResult.ParamNames.Add(Token.Data);
                   inc(spoint);
                   AParams[AParamCount]:=curtp;
                   inc(AParamCount);
                  end;
                  case nexttoken.id of
                   idcomma  :begin token:=readtoken;goto l2;end;
                   id2points:begin
                             token:=readtoken;
                             {getvartype;}
                             token:=Readtoken;
                             if madd=false then
                             for k:=1 to st do
                              Tidentlistitem(
                              Variables.Objects[il[k]]).VType:=token.data;
                             if nexttoken.id<>idclosebracket then
                              getdelimeter;
                             end;
                   else error(LoadStr(meth_decerr));
                  end;
                 end;
  idclosebracket:exit;
  id_var        :begin token:=readtoken;curtp:=1;goto l2;end;
  id_const      :begin token:=readtoken;goto l2;end;
 else Error(LoadStr(bad_methparam));
 end;
goto l1;
end;
{--}
begin

 getidentifier;ProcName:=token.data;
 while nexttoken.id=idpoint do begin
  token:=readtoken;
  getidentifier;
  ProcName:=ProcName+'.'+token.data;
 end;
 Madd:=varidbyname(ProcName,ident);
 AParamCount:=0;
 if madd=false then begin
  k:=IDLabels.newitem;
  Result:=AddVars(ProcName,k);
  EResult:=tidentlistitem(Variables.Objects[Result]);
 end;
 if nexttoken.id=idopenbracket then begin
  getopenbracket;
  getprocparams;
  getclosebracket;
 end;
if Madd=false then
 begin
   with EResult do begin
   Params:=AParams;
   ParCount:=AParamCount;
   IdentType:=1;
  end;
  if nexttoken.id=id2points then begin
   Token:=ReadToken;
   getvartype; {function result type}

   EResult.VType:=token.data;
  end;
   AddVars(ProcName+coProcResult,spoint);
   inc(spoint);
 end;
  varidbyname(ProcName,Result);
  getdelimeter;
  getafterproc;
end;
{------------------------------}
procedure THalCompiler.getafterproc;
var i:integer;
    s:boolean;
label l1;
begin
s:=true;
l1:
i:=Nexttoken.id;
if (i=id_virtual) or (i=id_override) or (i=id_dynamic) or (i=id_abstract)
then begin
      Token:=Readtoken;
      getdelimeter;
      if s then begin
       s:=false;
       goto l1;
      end;
     end;
end;
{------------------------------}
procedure THalCompiler.getclassdefblock(b:boolean);
label l1;
var i:integer;
begin
 l1:
 case nexttoken.id of
  ididentifier :begin getmyvariables(b);end;
  id_property  :begin Error(LoadStr(no_props));end;
  id_procedure,id_function :begin
                 Token:=ReadToken;
                 getprocdef(false);
                end;
 else exit;
 end;
 goto l1;
end;
{------------------------------}
procedure THalCompiler.getclassdef(const cname:string);
var ParName:String;
label l1;
begin
ParName:='';
IdentPrefix:=cname+'.';
 { get parent if any }
 if NextToken.ID=idopenbracket then
 begin
  Token:=ReadToken;
  Token:=ReadToken;
  if Token.ID<>ididentifier then
  Error(LoadStr(need_par));
  ParName:=Token.Data;
  Token:=ReadToken;
  if Token.ID<>idclosebracket then
   Error(LoadStr(clbr_exp));
 end;
AddObjectType(cname,ParName);
FLastClassType:=ParName;
 l1:
 case NextToken.id of
  iddelimeter:begin
              Token:=Readtoken;
              IdentPrefix:='';
              exit;
              end;
  id_private,id_public,id_published,id_protected:
               begin
                Token:=ReadToken;
                getclassdefblock(true);
               end;
  id_end      :begin
                Token:=ReadToken;
                getdelimeter;
                IdentPrefix:='';
                exit;
               end;
  else getclassdefblock(true);
 end;
 goto l1;
end;
{------------------------------}
procedure THalCompiler.gettypeidentifier;
var s:string;
begin
 Token:=ReadToken;
 S:=Token.Data;
 Token:=ReadToken;
 if Token.ID<>idequal then
  Error(FMTLoadStr(err_decl,[s]));
 Token:=ReadToken;
 Case Token.ID of
  id_record,id_class:getclassdef(s);
 else Error(LoadStr(only_class));
 end;
end;
{------------------------------}
procedure THalCompiler.getvartype;
begin
Token:=ReadToken;
{if ((Token.id>=idtypebase) and (token.id<=idtypeend))
or ((token.id=ididentifier) and
(RegisteredClassName(token.data)))
then begin exit;end;
Error('Type identifier expected');}
end;
{------------------------------}
procedure THalCompiler.getmyvariables(b:boolean);
label l6,l5;
var
    vcurent,ad,mid:integer;
    vc:array[0..200] of tidentlistitem;
    s:string;
    ident:tidentlistitem;
    tob:tobjitem;
    ob:boolean;
begin
 ob:=b;
 l6:
 vcurent:=0;
 l5:b:=ob;
 Token:=ReadToken;
 if Token.ID<>idIdentifier then
 Error(LoadStr(varname_exp));
 s:=Token.Data;
 if varbyname(s,ident) then Error(FmtLoadStr(var_already,[s]));

 if b then begin
  If mobjects.objbyname(s,tob) then b:=false;
 If B then begin
  mid:=AddVars(s,spoint);
  vc[vcurent]:=tidentlistitem(Variables.Objects[mid]);
  inc(vcurent);
  inc(spoint);
 end;end;
 Token:=ReadToken;
 case Token.ID of
 idcomma:goto l5;
 id2points:begin
           getvartype;
           if b then
           for ad:=0 to vcurent-1 do vc[ad].VType:=token.data;
           getdelimeter;
           if (NextToken.ID<=idReservedEnd) and
           (NextToken.ID>=idReservedBase)
           then exit
           else goto l6;
           end;
 else Error(LoadStr(bad_varblock));
 end;
end;
{------------------------------}
procedure THalCompiler.getprocbody(i:integer);
var oidentpref:string;
    l1:integer;
begin
if nexttoken.id=id_forward then
begin token:=readtoken;getdelimeter;exit;end;
oidentpref:=identprefix;
identprefix:=variables.strings[i]+'.';

getdeclarations;
l1:=(variables.objects[i] as tidentlistitem).id;
IDLabels.setplace(l1,prog.count);
IDLabels.SetReference(l1);

getoperatorblock;
getdelimeter;
identprefix:=oidentpref;
putcode(ocreturn,0,0);
end;
{------------------------------}
procedure THalCompiler.getdeclarations;
label l8,l3,l6;
var s:string;
    i,k:integer;
    ident:tidentlistitem;
begin
l8:
case nexttoken.id of
id_uses:begin
         Token:=ReadToken;
         l6:
         getidentifier;
         UseUnit(token.data);
         Token:=ReadToken;
         if token.id=idcomma then goto l6 else goto l8;
        end;
id_interface,id_implement:begin
                           Token:=ReadToken;
                           goto l8;
                          end;
id_begin,id_end:begin end;
id_procedure,id_function :begin
                           Token:=ReadToken;
                           getprocbody(getprocdef(true));
                           goto l8;
                          end;
id_label: begin
          Token:=ReadToken;
          l3:
          Token:=ReadToken;
          if Token.ID<>idIdentifier then
          Error(LoadStr(labname_exp));
          S:=Token.Data;
          if labels.idbyname(s,i) then
          Error(FMTLoadStr(label_already,[s]));
          k:=IDLabels.newitem;
          labels.additem(s,k);
          Token:=ReadToken;
          Case Token.Id of
          idDelimeter:goto l8;
          idComma:goto l3;
          else Error(LoadStr(delim_or_coma));
          end;
          end;
id_type:  begin
           Token:=ReadToken;
           while NextToken.ID=ididentifier do
            gettypeidentifier;
           goto l8;
          end;
id_var:   begin
          token:=readtoken;
          getmyvariables(true);
          goto l8;
          end;
{}
else Error(LoadStr(err_declpart));
end;
end;
{------------------------------}
procedure THalCompiler.getoperatorcoma;
begin
getoperator;
if (nexttoken.id=iddelimeter) or
((token.id<>id2points) and (nexttoken.id<>id_end) and (nexttoken.id<>id_until))
then getdelimeter;
end;
{------------------------------}
procedure THalCompiler.getwhileoperator;
var a,i,l2,l1:integer;
begin
Token:=ReadToken;
l1:=IDLabels.newitem;
IDLabels.SetReference(l1);
IDLabels.setplace(l1,prog.count);
l2:=IDLabels.newitem;
IDLabels.SetReference(l2);
i:=spoint;
a:=iexpression;
spoint:=i;
putcode(ocif,a,l2);
token:=readtoken;if Token.id<>id_do then Error(LoadStr(do_exp));
getoperator;
putcode(ocgoto,l1,0);
IDLabels.setplace(l2,prog.count);
end;
{------------------------------}
procedure THalCompiler.getrepeatoperator;
var i,a,l1:integer;
begin
token:=readtoken;
l1:=IDLabels.newitem;
IDLabels.SetReference(l1);
IDLabels.setplace(l1,prog.count);
while (nexttoken.id<>id_until) and (nexttoken.id<>idendoffile) do getoperatorcoma;
token:=readtoken;if token.id<>id_until then Error(LoadStr(until_exp));
i:=spoint;
a:=iexpression;
spoint:=i;
putcode(ocif,a,l1);
end;
{------------------------------}
procedure THalCompiler.getforoperator;
var mi,a,l,l2,l1,vid,i:integer;
    ident:tidentlistitem;
begin
token:=ReadToken;
token:=readtoken;if token.id<>ididentifier then Error(LoadStr(varname_exp));

if varbyname(token.data,ident)=false then
Error(FmtLoadStr(var_NotDef,[token.data]));
vid:=ident.id;
token:=readtoken;
if token.id<>id2points then Error(LoadStr(p2_exp));
token:=readtoken;if token.id<>idequal then Error(LoadStr(eq_exp));
i:=myexpress;
putcode(ocmov,vid,i);

token:=readtoken;
l1:=IDLabels.newitem;
IDLabels.SetReference(l1);
l2:=IDLabels.newitem;
IDLabels.SetReference(l2);
IDLabels.setplace(l1,prog.count);
l:=token.id;
{}
mi:=spoint;
a:=iexpression;
putcode(ocmov,spoint,vid);
token:=readtoken;if token.id<>id_do then Error(LoadStr(do_exp));
case l of
id_to    :begin
          putcode(oclessequal,spoint,a);
          putcode(ocif,spoint,l2);
          spoint:=mi;
          getoperator;
          putcode(ocincvar,vid,0);
          end;
id_downto:begin
          putcode(ocgreaterequal,spoint,a);
          putcode(ocif,spoint,l2);
          spoint:=mi;
          getoperator;
          putcode(ocdecvar,vid,0);
          end;
else Error(LoadStr(down_to_exp));
end;
putcode(ocgoto,l1,0);
IDLabels.setplace(l2,prog.count);
end;
{------------------------------}
procedure THalCompiler.getifoperator;
var i,a:integer;
    l1,l2:integer;
begin
Token:=ReadToken;
i:=spoint;
a:=iexpression;
spoint:=i;
l1:=IDLabels.newitem;
IDLabels.SetReference(l1);
putcode(ocif,a,l1);
token:=readtoken;
if token.id<>id_then then Error(LoadStr(then_exp));
getoperator;
if nexttoken.id=iddelimeter then begin
IDLabels.setplace(l1,prog.count);
exit;end;

token:=readtoken;
if token.id<>id_else then Error(LoadStr(else_exp));
l2:=IDLabels.newitem;
IDLabels.SetReference(l2);
putcode(ocgoto,l2,0);
IDLabels.setplace(l1,prog.count);
getoperator;
IDLabels.setplace(l2,prog.count);
end;
{------------------------------}
function THalCompiler.getvar:integer;
var t:tidentlistitem;
begin
token:=readtoken;
if token.id<>ididentifier then Error(LoadStr(varname_exp));
if varbyname(token.data,t)=false then
Error(FmtLoadStr(var_NotDef,[token.data]));
result:=t.id;
Lastvar:=t;
end;
{------------------------------}
procedure THalCompiler.getoperator;
var a1,a2,l,a,i:integer;
    s:string;
label l23;
begin
l:=NextToken.Id;
case l of
id_begin:getoperatorblock;
id_if:getifoperator;
id_while:getwhileoperator;
id_repeat:getrepeatoperator;
id_for:getforoperator;
id_Goto:  begin
           Token:=ReadToken;
           Token:=ReadToken;
           if Token.ID<>idIdentifier then Error(LoadStr(labname_exp));
           S:=Token.Data;
           if Labels.IDByName(s,i)=false
            then Error(FmtLoadStr(lab_notdef,[s]));
           IDLabels.SetReference(i);
           putcode(ocgoto,i,0);
          end;
idIdentifier:begin
             Token:=ReadToken;
             s:=Token.Data;
              if nexttoken.id=idpoint then begin
               findobject(s);
              end;
             while nexttoken.id=idpoint do begin
             token:=readtoken;
             token:=readtoken;
             if token.id<>ididentifier then Error(LoadStr(id_expected));
             s:=s+'.'+token.data;
             end;
             case nexttoken.id of
             idopenbracket,idsqopenbracket:l23:getuserfunction(s,true,true);
             id2points:begin Token:=ReadToken;
                       case nexttoken.id of
                       idequal:begin
                       Token:=ReadToken;
                       equaldispath(s);
                       end;
                       else labeldispatch(s);
                       end;
                       end;
             else goto l23;
             end;
             end;
else Error(FmtLoadStr(unkn_id,[nexttoken.data]));
{}
end;
end;
{------------------------------}
procedure THalCompiler.getoperatorblock;
begin
Token:=ReadToken;
if token.id=id_end then exit;
if token.id<>id_begin then Error(LoadStr(begin_expected));
while (nexttoken.id<>idEndOFFile) and (nexttoken.id<>id_end) do getoperatorcoma;
token:=ReadToken;
if token.id<>id_end then Error(LoadStr(end_expected));
end;
{------------------------------}
procedure THalCompiler.compileprogram;
var i:integer;
begin
 i:=IDLabels.newitem;
 IDLabels.SetReference(i);
 putcode(ocgoto,i,0);
 getprogramname;
 getdeclarations;
 IDLabels.setplace(i,prog.count);
 getoperatorblock;
 Token:=ReadToken;
 putcode(ochalt,0,0);
end;
{-----------------------------------------------------------------------}
procedure THalCompiler.writeprogram;
{-----------}
procedure writeconsts;
var i:integer;
    v:variant;
begin
outint(fout,consts.count);
for i:=0 to consts.count-1 do begin
v:=tconstitem(consts.items[i]).data;
case Vartype(v) of
varBoolean                     :begin outbyte(fout,wboolean);outboolean(fout,v);end;
varSingle,varDouble            :begin outbyte(fout,wdouble);outdouble(fout,v);end;
varSmallint,varbyte,varinteger :begin outbyte(fout,winteger);outint(fout,v);end;
varstring                      :begin outbyte(fout,wstring);outstring(fout,v);end;
else Error(LoadStr(unexp_writer));
end;end;
end;
{-----------}
procedure writelabels;
var i,c:integer;
begin
outint(fout,idlabels.count);
for i:=0 to idlabels.count-1 do begin
c:=tidlabelitem(idlabels.items[i]).Place;
outint(fout,c);
end;
end;
{-----------}
procedure writeprog;
var i:integer;
    v:tprogitem;
begin
outint(fout,prog.count);
for i:=0 to prog.count-1 do begin
v:=tprogitem(prog.items[i]);
outint(fout,v.a);
outint(fout,v.b);
outint(fout,v.opcode);
end;
end;
{-----------}
procedure writeheader;
begin
outstring(fout,binheader);
end;
{-----------}
begin
writeheader;
writeconsts;
writelabels;
writeprog;
end;
{------------------------------}
constructor THalCompiler.Create(afin:TMemoryStream;afout:TStream);
var t:integer;
begin
inherited create;myinit;fout:=afout;
{curread:=0;spoint:=0;}
inbuf:=afin.memory;
numread:=afin.size;
 {$IFDEF DEMO}
  If not DelphiRunning then begin
     ShowMessage(
     'You should register your copy of HAL'
     );
    Application.Terminate;
  end;
 {$ENDIF}
end;
{------------------------------}
destructor THalCompiler.Destroy;
begin
mydone;
inherited destroy;
end;
{------------------------------}
procedure THalCompiler.compile;
begin
compileprogram;
if idlabels.existlabels=false then Error(LoadStr(linker_error));
WriteProgram;
end;
{----------------------------------------}
procedure TBinProg.Error(const s:string);
begin
raise ECompilerError.Create(s);
end;
{------------------------------}
constructor TBinProg.Create(Fin:TStream);
{--}
procedure readconsts;
var i:integer;
    b:byte;
    v:variant;
begin
constsize:=getint(fin);
getmem(c,constsize*sizeof(variant));
fillchar(c^,constsize*sizeof(variant),0);
for i:=0 to constsize-1 do begin
b:=getbyte(fin);
case b of
wInteger:v:=getint(fin);
wDouble :v:=getdouble(fin);
wString :v:=getstring(fin);
WBoolean:v:=getboolean(fin);
else Error('Unknown reader type');
end;
c^[i]:=v;
end;
end;
{--}
procedure readprogram;
var i:integer;
begin
progsize:=getint(fin);
getmem(prog,progsize*sizeof(tbinprogitem));
for i:=0 to progsize-1 do begin
prog^[i].a:=getint(fin);
prog^[i].b:=getint(fin);
prog^[i].opcode:=getint(fin);
end;
end;
{--}
procedure readlabels;
var i:integer;
begin
labelsize:=getint(fin);
getmem(l,labelsize*sizeof(integer));
for i:=0 to labelsize-1 do l^[i]:=getint(fin);
end;
{--}
var i:integer;
begin
inherited create;
if getstring(Fin)<>binheader then Error('Bad header');
stacksize:=1000;
getmem(s,stacksize*sizeof(variant));
fillchar(s^,stacksize*sizeof(variant),0);
readconsts;
readlabels;
readprogram;
end;
{------------------------------}
function TBinProg.dispatch(opcode,a,b:integer):boolean;
var sa:string;
    vsize,ik,code:integer;
    mys:string;
    i:double;
    a1,a2:integer;
    ma:array[0..maxparams] of variant;
    mint:array[0..maxparams] of integer;
    mb:array[0..maxparams] of tvarrec;
    myfun:TFunListItem;

begin
inc(curpos);Result:=true;
case opcode of
ocincvar:begin S^[a]:=S^[a]+1;end;
ocdecvar:begin S^[a]:=S^[a]-1;end;
ocAdd        :S^[A]:=s^[a]+s^[b];
ocSub        :s^[a]:=s^[a]-s^[b];
ocMul        :s^[a]:=s^[a]*s^[b];
ocDiv        :s^[a]:=s^[a] div s^[b];
ocMod        :s^[a]:=s^[a] mod s^[b];
ocSlash      :s^[a]:=s^[a] / s^[b];
ocShl        :s^[a]:=s^[a] shl s^[b];
ocShr        :s^[a]:=s^[a] shr s^[b];
ocNot        :s^[a]:=not s^[a];
ocOr         :s^[a]:=s^[a] Or s^[b];
ocXor        :s^[a]:=s^[a] Xor s^[b];
ocAnd        :s^[a]:=s^[a] And s^[b];
ocGreaterEqual:s^[a]:=s^[a] >= s^[b];
ocEqual       :begin
                S^[a]:=(vartype(s^[a])=vartype(s^[b])) and (s^[a]=s^[b]);
               end;
ocLessEqual   :s^[a]:=s^[a] <= s^[b];
ocNotEqual    :begin
                s^[a]:=(vartype(s^[a])<>vartype(s^[b])) or (s^[a] <> s^[b]);
               end;
ocGreater     :s^[a]:=s^[a] > s^[b];
ocLess        :s^[a]:=s^[a] < s^[b];
ocNeg         :s^[a]:=-s^[a];
ocGoto        :curpos:=L^[a];
ocIF          :if s^[a]=false then curpos:=L^[B];
ocmov         :S^[a]:=S^[b];
ocselffromvar :begin
               lastobject:=tobject(VarToObj(S^[a]));
               end;
ocloadextvar  :begin
               { a-varid
                 b-funid
                 snum:=prog^[curpos].a
                 }
              a1:=prog^[curpos].a;
              inc(curpos);
              S^[a1]:=(DynaVars.Objects[b] as TDynaVarItem).GetVar(a);
              end;
occall        :begin
                CallStack[CallMax]:=curpos;
                inc(CallMax);
                Curpos:=L^[a];
                exit;
               end;
ocvarraycreate:begin
               S^[a]:=VarArrayCreate([0,b],varVariant);
               end;
ocsetvarray:   begin
               a1:=prog^[curpos].a;
               S^[a1][a]:=S^[b];
               inc(curpos);
               end;
ochalt:       begin
              Result:=false;
              exit;
              end;
ocreturn      :begin
                IF CallMax=0 then begin Result:=false;exit;end;
                dec(callmax);
                Curpos:=CallStack[callmax];
                exit;
               end;
ocstoreextvar :begin
              a1:=prog^[curpos].a;
              inc(curpos);
              (DynaVars.Objects[b] as TDynaVarItem).SetVar(a,S^[a1]);
               end;
ocloadconst   :begin S^[a]:=C^[b];end;
ocsetself     :begin lastobject:=tobject(a);end;

ocextfun      :begin
              {if not assigned(funs) then Error('User Functions not defined');}
              myfun:=tfunlistitem(Funs.Objects[b]);
              if not assigned(myfun) then Error('Unknown user Function');
              if myfun.parcount>0 then begin
              ik:=0;
              while ik<=myfun.parcount-1 do begin
              mint[ik]:=prog^[curpos].opcode;
              mint[ik+1]:=prog^[curpos].a;
              mint[ik+2]:=prog^[curpos].b;
              inc(curpos);
              inc(ik,3);
              end;
              end;
              if myfun.parcount>0 then begin
              for ik:=0 to myfun.parcount-1 do
              ma[ik]:=S^[mint[ik]];
              S^[a]:=myfun.ProcAddr(lastobject,slice(ma,myfun.parcount));
              for ik:=0 to myfun.parcount-1 do if myfun.params[ik]=1 then
              S^[mint[ik]]:=ma[ik];{}
              end else S^[a]:=myfun.ProcAddr(lastobject,ma);
              end;
else Error('Invalid opcode');
end;end;
{------------------------------}
procedure TBinProg.Run;
begin
RunFrom(0);
end;
{------------------------------}
procedure TBinProg.RunFrom(Acurpos:Integer);
begin
curpos:=ACurPos;
while
dispatch(prog^[curpos].opcode,prog^[curpos].a,prog^[curpos].b)
do begin end;
end;
{------------------------------}
destructor  TBinProg.Destroy;
begin
if assigned(prog) then freemem(prog,progsize*sizeof(tbinprogitem));
if assigned(c) then freemem(c,constsize*sizeof(variant));
if assigned(l) then freemem(l,labelsize*sizeof(integer));
if assigned(s) then freemem(s,stacksize*sizeof(variant));
inherited destroy;
end;
{------------------------------}
procedure thalRuner.AddObjectbyRef(const objname:string;tobj:tobject);
begin
 AddObject(ObjName,tobj.ClassName,tobj);
end;
{------------------------------}
procedure thalRuner.AddObject(const objname,objtype:string;tobj:tobject);
 var t:tobjitem;
begin
 t:=tobjitem.create;
 t.objtype:=ansiuppercase(objtype);
 t.tobj:=tobj;
 myobjects.addobject(ansiuppercase(objname),t);
end;
{-------------------}
constructor THalRuner.Create(T:TMemoryStream);
begin
 inherited Create;
 instream:=t;
 myobjects:=Tobjcollect.create;
 TempVarList:=TIdentList.Create;
end;
{-------------------}
procedure THalRuner.Run;
var FHalCompiler:THalCompiler;
begin
if compiled=false then begin
try
mymem:=TmemoryStream.create;
FHalCompiler:=THalCompiler.Create(instream,mymem);
FHalCOmpiler.HALOWNER:=HalOwner;
FHalCompiler.mobjects:=myobjects;
{try}
FHalCompiler.compile;
finally
 FLastClassType:=FHalCompiler.FLastClassType;
 if loadtemp then TempVarList.Assign(FHalCompiler.Variables);
 FHalCompiler.free;
end;
mymem.seek(0,0);
FBinProg:=TBinProg.Create(mymem);
compiled:=true;
end;
if dontrun then exit;
FBinProg.Run;
end;
{-------------------}
destructor  THalRuner.Destroy;
begin
 if compiled then begin
  mymem.free;
  FBinProg.Free;
  myobjects.free;
 end;
 TempVarList.Free;
 inherited;
end;
{----------------------------------------------}
end.
